import { CommonModule } from '@angular/common';
import { takeUntil } from 'rxjs/operators';
import { EventEmitter, Injectable, ChangeDetectorRef, Directive, Inject, Input, SimpleChange, ViewContainerRef, NgModule } from '@angular/core';
import { BehaviorSubject, isObservable, Observable, ReplaySubject, Subject } from 'rxjs';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const defaultNgxBindIOConfig = {
    debug: false
};
/** @type {?} */
const NGX_BIND_IO_CONFIG = 'NgxBindIOConfig';

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} component
 * @return {?}
 */
function getBindIOMetadata(component) {
    /** @type {?} */
    const emptyMetadata = {
        asHost: {
            subjects: {}
        },
        asInner: {
            subscriptions: {},
            manualInputs: undefined,
            manualOutputs: undefined
        }
    };
    if (component) {
        if (!component['__bindIO__']) {
            ((/** @type {?} */ (component['__bindIO__']))) = emptyMetadata;
        }
        return (/** @type {?} */ (component['__bindIO__']));
    }
    return emptyMetadata;
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} component
 * @param {?} rule
 * @param {?=} maxLevel
 * @return {?}
 */
function collectKeys(component, rule, maxLevel) {
    if (maxLevel !== undefined) {
        maxLevel--;
        if (maxLevel === 0) {
            return [];
        }
    }
    /** @type {?} */
    let keys = component ? Object.keys(component).filter(propName => rule(component, propName)) : [];
    if (component && component.__proto__) {
        keys = [...keys, ...collectKeys(component.__proto__, rule, maxLevel)];
    }
    return keys;
}
/**
 * @param {?} directive
 * @param {?} existsKeys
 * @return {?}
 */
function removeKeysUsedInAttributes(directive, existsKeys) {
    /** @type {?} */
    const viewContainerRef = directive.viewContainerRef;
    if (viewContainerRef &&
        viewContainerRef.element &&
        viewContainerRef.element.nativeElement &&
        viewContainerRef.element.nativeElement.attributes) {
        /** @type {?} */
        const attributes = (/** @type {?} */ (viewContainerRef.element.nativeElement.attributes));
        return existsKeys.filter(existKey => {
            /** @type {?} */
            let founded = false;
            for (let i = 0; i < attributes.length; i++) {
                /** @type {?} */
                const name = attributes
                    .item(i)
                    .name.replace(new RegExp('-', 'g'), '')
                    .toUpperCase();
                if (existKey.replace(new RegExp('-', 'g'), '').toUpperCase() === name) {
                    founded = true;
                }
            }
            return !founded;
        });
    }
    return existsKeys;
}
/**
 * @param {?} directive
 * @param {?} existsKeys
 * @return {?}
 */
function removeKeysManualBindedOutputs(directive, existsKeys) {
    return existsKeys.filter(innerKey => {
        return ((getBindIOMetadata(directive.innerComponent).asInner.manualOutputs
            ? Object.keys(getBindIOMetadata(directive.innerComponent).asInner.manualOutputs)
            : []).filter(outputName => outputName.toUpperCase() === innerKey.toUpperCase()).length === 0);
    });
}
/**
 * @param {?} directive
 * @param {?} existsKeys
 * @return {?}
 */
function removeKeysManualBindedInputs(directive, existsKeys) {
    return existsKeys.filter(innerKey => {
        return ((getBindIOMetadata(directive.innerComponent).asInner.manualInputs
            ? Object.keys(getBindIOMetadata(directive.innerComponent).asInner.manualInputs)
            : []).filter(inputName => inputName.toUpperCase() === innerKey.toUpperCase()).length === 0);
    });
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/** @type {?} */
const values = new WeakMap();
/**
 * @param {?} instance
 * @return {?}
 */
function valueMap(instance) {
    /** @type {?} */
    let _valueMap = values.get(instance);
    if (!_valueMap) {
        _valueMap = new Map();
        values.set(instance, _valueMap);
    }
    return _valueMap;
}
/**
 * @param {?} target
 * @param {?} propertyKey
 * @param {?=} callback
 * @return {?}
 */
function redefineSimpleProperty(target, propertyKey, callback) {
    Object.defineProperty(target, propertyKey, {
        configurable: true,
        enumerable: true,
        /**
         * @param {?} value
         * @return {?}
         */
        set(value) {
            valueMap(this).set(propertyKey, value);
            if (callback) {
                callback(value);
            }
        },
        /**
         * @return {?}
         */
        get() {
            return valueMap(this).get(propertyKey);
        }
    });
}
/**
 * @param {?} target
 * @param {?} propertyKey
 * @param {?} descriptor
 * @param {?=} callback
 * @return {?}
 */
function redefineAccessorProperty(target, propertyKey, descriptor, callback) {
    Object.defineProperty(target, propertyKey, {
        configurable: descriptor.configurable,
        enumerable: descriptor.enumerable,
        /**
         * @param {?} value
         * @return {?}
         */
        set(value) {
            if (!descriptor.set) {
                throw new Error(`Property ${propertyKey} doesn't have a setter and cannot be written`);
            }
            descriptor.set.call(this, value);
            if (!descriptor.get) {
                callback(value);
            }
            else {
                callback(descriptor.get());
            }
        },
        /**
         * @return {?}
         */
        get() {
            if (!descriptor.get) {
                throw new Error(`Property ${propertyKey} doesn't have a getter and cannot be read`);
            }
            return descriptor.get.call(this);
        }
    });
}
/**
 * @param {?} target
 * @param {?} propertyKey
 * @return {?}
 */
function getPropDescriptor(target, propertyKey) {
    /** @type {?} */
    let originalDescriptor = Reflect.getOwnPropertyDescriptor(target, propertyKey);
    /** @type {?} */
    let setter = originalDescriptor && originalDescriptor.set;
    /** @type {?} */
    let getter = originalDescriptor && originalDescriptor.get;
    /** @type {?} */
    const value = originalDescriptor ? originalDescriptor.value : target[propertyKey];
    if (!setter || ((/** @type {?} */ (target))).__lookupSetter__(propertyKey)) {
        setter = ((/** @type {?} */ (target))).__lookupSetter__(propertyKey);
    }
    if (!getter || ((/** @type {?} */ (target))).__lookupGetter__(propertyKey)) {
        getter = ((/** @type {?} */ (target))).__lookupGetter__(propertyKey);
    }
    if (!originalDescriptor && setter && getter) {
        originalDescriptor = {
            configurable: true,
            enumerable: true,
            set: setter.bind(target),
            get: getter.bind(target)
        };
    }
    return {
        originalDescriptor: originalDescriptor,
        value,
        setter,
        getter
    };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
/**
 * @param {?} functionToCheck
 * @return {?}
 */
function isFunction(functionToCheck) {
    return functionToCheck && {}.toString.call(functionToCheck) === '[object Function]';
}
/**
 * @param {?} derivedCtor
 * @param {?} baseCtors
 * @return {?}
 */
function applyMixins(derivedCtor, baseCtors) {
    baseCtors.forEach(baseCtor => {
        Object.getOwnPropertyNames(baseCtor.prototype).forEach(name => {
            derivedCtor.prototype[name] = baseCtor.prototype[name];
        });
    });
    return derivedCtor;
}
/**
 * @param {?} baseCtors
 * @return {?}
 */
function Mixin(baseCtors) {
    return function (derivedCtor) {
        baseCtors.forEach(baseCtor => {
            Object.getOwnPropertyNames(baseCtor.prototype).forEach(name => {
                derivedCtor.prototype[name] = baseCtor.prototype[name];
            });
        });
    };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindInputsService {
    /**
     * BindInputs Directive
     * @param {?} directive
     * @return {?}
     */
    bindInputs(directive) {
        /** @type {?} */
        const inputs = directive.inputs;
        const { includeInputs, excludeInputs } = this.getIncludesAndExcludes(directive);
        inputs.hostKeys
            .filter(hostKey => (includeInputs.length === 0 && excludeInputs.indexOf(hostKey.toUpperCase()) === -1) ||
            (includeInputs.length !== 0 && includeInputs.indexOf(hostKey.toUpperCase()) !== -1))
            .forEach(hostKey => {
            inputs.innerKeys
                .filter(innerKey => (includeInputs.length === 0 && excludeInputs.indexOf(innerKey.toUpperCase()) === -1) ||
                (includeInputs.length !== 0 && includeInputs.indexOf(innerKey.toUpperCase()) !== -1))
                .forEach(innerKey => {
                if (this.checkInputToBind(directive, hostKey, innerKey)) {
                    this.bindInput(directive, hostKey, innerKey);
                }
            });
        });
    }
    /**
     * @param {?} directive
     * @return {?}
     */
    bindObservableInputs(directive) {
        /** @type {?} */
        const inputs = directive.inputs;
        const { includeInputs, excludeInputs } = this.getIncludesAndExcludes(directive);
        inputs.hostKeys
            .filter(hostKey => (includeInputs.length === 0 && excludeInputs.indexOf(hostKey.toUpperCase()) === -1) ||
            (includeInputs.length !== 0 && includeInputs.indexOf(hostKey.toUpperCase()) !== -1))
            .forEach(hostKey => {
            inputs.innerKeys
                .filter(innerKey => (includeInputs.length === 0 && excludeInputs.indexOf(innerKey.toUpperCase()) === -1) ||
                (includeInputs.length !== 0 && includeInputs.indexOf(innerKey.toUpperCase()) !== -1))
                .forEach(innerKey => {
                if (this.checkObservableInputToBind(directive, hostKey, innerKey)) {
                    this.bindObservableInput(directive, hostKey, innerKey);
                }
            });
        });
    }
    /**
     * Inputs
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkKeyNameToInputBind(directive, hostKey, innerKey) {
        return hostKey === innerKey && hostKey[0] !== '_';
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkInputToBind(directive, hostKey, innerKey) {
        /** @type {?} */
        const hostValue = getPropDescriptor(directive.hostComponent, hostKey).value || directive.hostComponent[hostKey];
        return (directive.usedInputs[hostKey] === undefined &&
            !isObservable(hostValue) &&
            this.checkKeyNameToInputBind(directive, hostKey, innerKey));
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    bindInput(directive, hostKey, innerKey) {
        directive.usedInputs[hostKey] = innerKey;
        /** @type {?} */
        const descriptor = getPropDescriptor(directive.hostComponent, hostKey);
        /** @type {?} */
        const currentValue = descriptor.value || directive.hostComponent[hostKey];
        if (!getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey]) {
            getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey] = new Subject();
            delete directive.hostComponent[hostKey];
            if (descriptor !== undefined && descriptor.setter !== undefined && descriptor.getter !== undefined) {
                redefineAccessorProperty(directive.hostComponent, hostKey, descriptor.originalDescriptor, (newValue) => getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey].next(newValue));
            }
            else {
                redefineSimpleProperty(directive.hostComponent, hostKey, (newValue) => getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey].next(newValue));
            }
        }
        getBindIOMetadata(directive.hostComponent)
            .asHost.subjects[hostKey].pipe(takeUntil(directive.destroyed$))
            .subscribe(value => directive.bindValue(innerKey, value));
        directive.hostComponent[hostKey] = currentValue;
    }
    /**
     * Observable Inputs
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkKeyNameToObservableInputBind(directive, hostKey, innerKey) {
        return hostKey === `${innerKey}$` && hostKey[0] !== '_';
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkObservableInputToBind(directive, hostKey, innerKey) {
        /** @type {?} */
        const hostValue = getPropDescriptor(directive.hostComponent, hostKey).value || directive.hostComponent[hostKey];
        return (directive.usedInputs[hostKey] === undefined &&
            isObservable(hostValue) &&
            this.checkKeyNameToObservableInputBind(directive, hostKey, innerKey));
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    bindObservableInput(directive, hostKey, innerKey) {
        directive.usedInputs[hostKey] = innerKey;
        /** @type {?} */
        const descriptor = getPropDescriptor(directive.hostComponent, hostKey);
        /** @type {?} */
        const isBehaviorSubject = descriptor.value instanceof BehaviorSubject;
        /** @type {?} */
        const currentValue = descriptor.value || directive.hostComponent[hostKey];
        /** @type {?} */
        let behaviorSubjectValue;
        if (descriptor.value instanceof BehaviorSubject) {
            behaviorSubjectValue = descriptor.value.getValue();
        }
        if (!getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey]) {
            getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey] = new Subject();
            delete directive.hostComponent[hostKey];
            if (descriptor !== undefined && descriptor.setter !== undefined && descriptor.getter !== undefined) {
                redefineAccessorProperty(directive.hostComponent, hostKey, descriptor.originalDescriptor, (newValue) => getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey].next(newValue));
            }
            else {
                redefineSimpleProperty(directive.hostComponent, hostKey, (newValue) => getBindIOMetadata(directive.hostComponent).asHost.subjects[hostKey].next(newValue));
            }
        }
        getBindIOMetadata(directive.hostComponent)
            .asHost.subjects[hostKey].pipe(takeUntil(directive.destroyed$))
            .subscribe(newValue => {
            if (newValue instanceof Observable) {
                if (getBindIOMetadata(directive.innerComponent).asInner.subscriptions[hostKey]) {
                    getBindIOMetadata(directive.innerComponent).asInner.subscriptions[hostKey].unsubscribe();
                }
                getBindIOMetadata(directive.innerComponent).asInner.subscriptions[hostKey] = newValue
                    .pipe(takeUntil(directive.destroyed$))
                    .subscribe(value => directive.bindValue(innerKey, value));
            }
        });
        try {
            directive.hostComponent[hostKey] = currentValue;
        }
        catch (error) { }
        if (isBehaviorSubject) {
            currentValue.next(behaviorSubjectValue);
        }
    }
    /**
     * Utils
     * @param {?} directive
     * @return {?}
     */
    getInputs(directive) {
        /** @type {?} */
        const foundedInputs = {
            hostKeys: collectKeys(directive.hostComponent, (cmp, hostKey) => !isFunction(getPropDescriptor(directive.innerComponent, hostKey).value) ||
                !isFunction(directive.innerComponent[hostKey]), 10),
            innerKeys: directive.innerComponent
                ? [
                    ...Object.keys(directive.innerComponent).filter(innerKey => !(getPropDescriptor(directive.innerComponent, innerKey).value instanceof EventEmitter ||
                        directive.innerComponent[innerKey] instanceof EventEmitter)),
                    ...collectKeys(directive.innerComponent.__proto__, (cmp, innerKey) => !(getPropDescriptor(cmp, innerKey).value instanceof EventEmitter), 10)
                ]
                : []
        };
        foundedInputs.hostKeys = [
            ...foundedInputs.hostKeys,
            ...foundedInputs.hostKeys
                .map(hostKey => hostKey + '$')
                .filter(hostKey => foundedInputs.hostKeys.indexOf(hostKey) === -1 &&
                (directive.hostComponent[hostKey] instanceof ReplaySubject ||
                    directive.hostComponent[hostKey] instanceof Subject ||
                    directive.hostComponent[hostKey] instanceof BehaviorSubject))
        ];
        foundedInputs.innerKeys = removeKeysManualBindedInputs(directive, removeKeysUsedInAttributes(directive, foundedInputs.innerKeys));
        foundedInputs.hostKeys = removeKeysUsedInAttributes(directive, foundedInputs.hostKeys);
        return foundedInputs;
    }
    /**
     * @param {?} directive
     * @return {?}
     */
    getIncludesAndExcludes(directive) {
        /** @type {?} */
        const exclude = Array.isArray(directive.excludeInputs) ? directive.excludeInputs : [directive.excludeInputs];
        /** @type {?} */
        const include = Array.isArray(directive.includeInputs) ? directive.includeInputs : [directive.includeInputs];
        /** @type {?} */
        const includeIO = !directive.includeIO
            ? []
            : Array.isArray(directive.includeIO)
                ? directive.includeIO
                : [directive.includeIO];
        /** @type {?} */
        const excludeIO = !directive.excludeIO
            ? []
            : Array.isArray(directive.excludeIO)
                ? directive.excludeIO
                : [directive.excludeIO];
        /** @type {?} */
        const excludeInputs = [...exclude, ...excludeIO].map(excludeKey => excludeKey.toUpperCase());
        /** @type {?} */
        const includeInputs = [...include, ...includeIO].map(includeKey => includeKey.toUpperCase());
        return { includeInputs, excludeInputs };
    }
}
NgxBindInputsService.decorators = [
    { type: Injectable }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindIODebugService {
    /**
     * @param {?} _ngxBindInputsService
     */
    constructor(_ngxBindInputsService) {
        this._ngxBindInputsService = _ngxBindInputsService;
    }
    /**
     * @param {?} directive
     * @param {?} debug
     * @return {?}
     */
    showDebugInfo(directive, debug) {
        /** @type {?} */
        let notExistsOutputs = [];
        /** @type {?} */
        let notExistsInputs = [];
        if (directive.outputs &&
            directive.innerComponent &&
            directive.innerComponent.__proto__ &&
            directive.innerComponent.__proto__.constructor &&
            directive.innerComponent.__proto__.constructor.ngBaseDef &&
            directive.innerComponent.__proto__.constructor.ngBaseDef.outputs) {
            /** @type {?} */
            const ngBaseDefOutputs = Object.keys(directive.innerComponent.__proto__.constructor.ngBaseDef.outputs);
            notExistsOutputs = removeKeysManualBindedOutputs(directive, removeKeysUsedInAttributes(directive, ngBaseDefOutputs)).filter(ngBaseDefOutput => directive.outputs.innerKeys.indexOf(ngBaseDefOutput) === -1);
        }
        if (directive.inputs &&
            directive.innerComponent &&
            directive.innerComponent.__proto__ &&
            directive.innerComponent.__proto__.constructor &&
            directive.innerComponent.__proto__.constructor.ngBaseDef &&
            directive.innerComponent.__proto__.constructor.ngBaseDef.inputs) {
            /** @type {?} */
            const ngBaseDefInputs = Object.keys(directive.innerComponent.__proto__.constructor.ngBaseDef.inputs);
            notExistsInputs = removeKeysManualBindedInputs(directive, removeKeysUsedInAttributes(directive, ngBaseDefInputs)).filter(ngBaseDefInput => directive.inputs.innerKeys.indexOf(ngBaseDefInput) === -1);
        }
        if (debug || notExistsOutputs.length > 0 || notExistsInputs.length > 0) {
            if (debug) {
                console.group('NgxBindIO: debug');
            }
            else {
                console.group('NgxBindIO: warning');
            }
            console.log('Host component:', directive.hostComponent.__proto__.constructor.name, directive.hostComponent);
            console.log('Inner component:', directive.innerComponent.__proto__.constructor.name, directive.innerComponent);
            if (directive.usedOutputs) {
                console.log('Outputs maping:', Object.keys(directive.usedOutputs).map(hostKey => `(${directive.usedOutputs[hostKey]})="${hostKey}($event)"`));
            }
            if (notExistsOutputs.length > 0) {
                console.log('Not initialized outputs:', notExistsOutputs);
                console.log('Not initialized outputs (text):', JSON.stringify(notExistsOutputs));
            }
            if (directive.usedInputs) {
                console.log('Inputs maping:', Object.keys(directive.usedInputs).map(hostKey => this._ngxBindInputsService.checkKeyNameToObservableInputBind(directive, hostKey, directive.usedInputs[hostKey])
                    ? `[${directive.usedInputs[hostKey]}]="${hostKey} | async"`
                    : `[${directive.usedInputs[hostKey]}]="${hostKey}"`));
            }
            if (notExistsInputs.length > 0) {
                console.log('Not initialized inputs:', notExistsInputs);
                console.log('Not initialized inputs (text):', JSON.stringify(notExistsInputs));
            }
            console.groupEnd();
        }
    }
}
NgxBindIODebugService.decorators = [
    { type: Injectable }
];
/** @nocollapse */
NgxBindIODebugService.ctorParameters = () => [
    { type: NgxBindInputsService }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindInputsDirective {
    /**
     * @param {?} viewContainerRef
     * @param {?} _ngxBindIOConfig
     * @param {?} _ngxBindInputsService
     * @param {?} _ngxBindIODebugService
     * @param {?} _ref
     */
    constructor(viewContainerRef, _ngxBindIOConfig, _ngxBindInputsService, _ngxBindIODebugService, _ref) {
        this.viewContainerRef = viewContainerRef;
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._ngxBindInputsService = _ngxBindInputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
        this._ref = _ref;
        this.excludeInputs = [];
        this.includeInputs = [];
        this.usedInputs = {};
        this.destroyed$ = new Subject();
        this.innerSimpleChanges = {};
    }
    /**
     * @param {?} simpleChanges
     * @return {?}
     */
    ngOnChanges(simpleChanges) {
        this.detectComponents();
    }
    /**
     * @return {?}
     */
    ngAfterContentInit() {
        this.detectComponents();
        this.bindAll();
    }
    /**
     * @return {?}
     */
    bindAll() {
        this.inputs = this._ngxBindInputsService.getInputs(this);
        this._ngxBindInputsService.bindInputs(this);
        this._ngxBindInputsService.bindObservableInputs(this);
        this._ngxBindIODebugService.showDebugInfo(this, this.debugIsActive());
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.destroyed$.next(true);
        this.destroyed$.complete();
    }
    /**
     * @param {?} innerKey
     * @param {?} value
     * @return {?}
     */
    bindValue(innerKey, value) {
        /** @type {?} */
        const previousValue = this.innerComponent[innerKey];
        this.innerComponent[innerKey] = value;
        this._ref.markForCheck();
        if (typeof this.innerComponent['ngOnChanges'] === 'function') {
            /** @type {?} */
            const simpleChange = new SimpleChange(previousValue, value, this.innerSimpleChanges[innerKey] === undefined);
            this.innerComponent['ngOnChanges']({ [innerKey]: simpleChange });
            this.innerSimpleChanges[innerKey] = simpleChange;
        }
    }
    /**
     * @return {?}
     */
    detectComponents() {
        if (!this.innerComponent && !this.hostComponent) {
            this.innerComponent = this.viewContainerRef['_data'].componentView.component;
            this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.context;
            if (this.hostComponent.$implicit !== undefined) {
                this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.component;
            }
        }
    }
    /**
     * @return {?}
     */
    debugIsActive() {
        return (this._ngxBindIOConfig.debug ||
            (this.bindInputs && this.bindInputs.debug) ||
            (localStorage && localStorage.getItem('debug_ngx-bind-io') === 'true')); // todo: remove on stable release
    }
}
BindInputsDirective.decorators = [
    { type: Directive, args: [{
                selector: '[bindInputs]'
            },] }
];
/** @nocollapse */
BindInputsDirective.ctorParameters = () => [
    { type: ViewContainerRef },
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: NgxBindInputsService },
    { type: NgxBindIODebugService },
    { type: ChangeDetectorRef }
];
BindInputsDirective.propDecorators = {
    bindInputs: [{ type: Input }],
    excludeInputs: [{ type: Input }],
    includeInputs: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindOutputsService {
    /**
     * BindOutputs Directive
     * @param {?} directive
     * @return {?}
     */
    bindOutputs(directive) {
        /** @type {?} */
        const outputs = directive.outputs;
        const { includeOutputs, excludeOutputs } = this.getIncludesAndExcludes(directive);
        outputs.hostKeys
            .filter(hostKey => (includeOutputs.length === 0 && excludeOutputs.indexOf(hostKey.toUpperCase()) === -1) ||
            (includeOutputs.length !== 0 && includeOutputs.indexOf(hostKey.toUpperCase()) !== -1))
            .forEach(hostKey => {
            outputs.innerKeys
                .filter(innerKey => (includeOutputs.length === 0 && excludeOutputs.indexOf(innerKey.toUpperCase()) === -1) ||
                (includeOutputs.length !== 0 && includeOutputs.indexOf(innerKey.toUpperCase()) !== -1))
                .forEach(innerKey => {
                if (this.checkOutputToBind(directive, hostKey, innerKey)) {
                    this.bindOutput(directive, hostKey, innerKey);
                }
            });
        });
    }
    /**
     * Outputs
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkKeyNameToOutputBind(directive, hostKey, innerKey) {
        /** @type {?} */
        const outputs = directive.outputs;
        /** @type {?} */
        const keyWithFirstUpperLetter = innerKey.length > 0 ? innerKey.charAt(0).toUpperCase() + innerKey.substr(1) : innerKey;
        return ((hostKey === `on${keyWithFirstUpperLetter}` &&
            outputs.hostKeys.indexOf(`on${keyWithFirstUpperLetter}Click`) === -1) ||
            hostKey === `on${keyWithFirstUpperLetter}Click`);
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    checkOutputToBind(directive, hostKey, innerKey) {
        /** @type {?} */
        const value = getPropDescriptor(directive.innerComponent, innerKey).value || directive.innerComponent[innerKey];
        return (directive.usedOutputs[hostKey] === undefined &&
            value instanceof EventEmitter &&
            this.checkKeyNameToOutputBind(directive, hostKey, innerKey));
    }
    /**
     * @param {?} directive
     * @param {?} hostKey
     * @param {?} innerKey
     * @return {?}
     */
    bindOutput(directive, hostKey, innerKey) {
        directive.usedOutputs[hostKey] = innerKey;
        directive.innerComponent[innerKey].subscribe(value => directive.hostComponent[hostKey](value));
    }
    /**
     * Utils
     * @param {?} directive
     * @return {?}
     */
    getOutputs(directive) {
        /** @type {?} */
        const foundedOutputs = {
            hostKeys: [
                ...Object.keys(directive.hostComponent).filter(hostKey => isFunction(directive.hostComponent[hostKey])),
                ...collectKeys(directive.hostComponent.__proto__, (cmp, hostKey) => isFunction(getPropDescriptor(cmp, hostKey).value), 10)
            ],
            innerKeys: directive.innerComponent
                ? [
                    ...Object.keys(directive.innerComponent).filter(innerKey => getPropDescriptor(directive.innerComponent, innerKey).value instanceof EventEmitter ||
                        directive.innerComponent[innerKey] instanceof EventEmitter),
                    ...collectKeys(directive.innerComponent.__proto__, (cmp, innerKey) => getPropDescriptor(cmp, innerKey).value instanceof EventEmitter, 10)
                ]
                : []
        };
        foundedOutputs.innerKeys = removeKeysManualBindedOutputs(directive, removeKeysUsedInAttributes(directive, foundedOutputs.innerKeys));
        foundedOutputs.hostKeys = removeKeysUsedInAttributes(directive, foundedOutputs.hostKeys);
        return foundedOutputs;
    }
    /**
     * @param {?} directive
     * @return {?}
     */
    getIncludesAndExcludes(directive) {
        /** @type {?} */
        const exclude = Array.isArray(directive.excludeOutputs) ? directive.excludeOutputs : [directive.excludeOutputs];
        /** @type {?} */
        const include = Array.isArray(directive.includeOutputs) ? directive.includeOutputs : [directive.includeOutputs];
        /** @type {?} */
        const includeIO = !directive.includeIO
            ? []
            : Array.isArray(directive.includeIO)
                ? directive.includeIO
                : [directive.includeIO];
        /** @type {?} */
        const excludeIO = !directive.excludeIO
            ? []
            : Array.isArray(directive.excludeIO)
                ? directive.excludeIO
                : [directive.excludeIO];
        /** @type {?} */
        const excludeOutputs = [...exclude, ...excludeIO].map(exludeKey => exludeKey.toUpperCase());
        /** @type {?} */
        const includeOutputs = [...include, ...includeIO].map(includeKey => includeKey.toUpperCase());
        return { includeOutputs, excludeOutputs };
    }
}
NgxBindOutputsService.decorators = [
    { type: Injectable }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindIODirective {
    /**
     * @param {?} viewContainerRef
     * @param {?} _ngxBindIOConfig
     * @param {?} _ngxBindInputsService
     * @param {?} _ngxBindOutputsService
     * @param {?} _ngxBindIODebugService
     * @param {?} _ref
     */
    constructor(viewContainerRef, _ngxBindIOConfig, _ngxBindInputsService, _ngxBindOutputsService, _ngxBindIODebugService, _ref) {
        this.viewContainerRef = viewContainerRef;
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._ngxBindInputsService = _ngxBindInputsService;
        this._ngxBindOutputsService = _ngxBindOutputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
        this._ref = _ref;
        this.excludeIO = [];
        this.includeIO = [];
        this.excludeInputs = [];
        this.includeInputs = [];
        this.excludeOutputs = [];
        this.includeOutputs = [];
        this.inputs = {
            innerKeys: [],
            hostKeys: []
        };
        this.usedInputs = {};
        this.usedOutputs = {};
        this.destroyed$ = new Subject();
        this.innerSimpleChanges = {};
    }
    /**
     * @param {?} simpleChanges
     * @return {?}
     */
    ngOnChanges(simpleChanges) {
        this.detectComponents();
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        this.detectComponents();
        this.bindAll();
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.destroyed$.next(true);
        this.destroyed$.complete();
    }
    /**
     * @param {?} innerKey
     * @param {?} value
     * @return {?}
     */
    bindValue(innerKey, value) {
        /** @type {?} */
        const previousValue = this.innerComponent[innerKey];
        this.innerComponent[innerKey] = value;
        this._ref.markForCheck();
        if (typeof this.innerComponent['ngOnChanges'] === 'function') {
            /** @type {?} */
            const simpleChange = new SimpleChange(previousValue, value, this.innerSimpleChanges[innerKey] === undefined);
            this.innerComponent['ngOnChanges']({ [innerKey]: simpleChange });
            this.innerSimpleChanges[innerKey] = simpleChange;
        }
    }
    /**
     * @return {?}
     */
    bindAll() {
        this.inputs = this._ngxBindInputsService.getInputs(this);
        this._ngxBindInputsService.bindInputs(this);
        this._ngxBindInputsService.bindObservableInputs(this);
        this.outputs = this._ngxBindOutputsService.getOutputs(this);
        this._ngxBindOutputsService.bindOutputs(this);
        this._ngxBindIODebugService.showDebugInfo(this, this.debugIsActive());
    }
    /**
     * @return {?}
     */
    detectComponents() {
        if (!this.innerComponent && !this.hostComponent) {
            this.innerComponent = this.viewContainerRef['_data'].componentView.component;
            this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.context;
            if (this.hostComponent.$implicit !== undefined) {
                this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.component;
            }
            getBindIOMetadata(this.innerComponent).asInner.manualOutputs = {};
            Object.keys(this.innerComponent)
                .filter(innerKey => this.innerComponent[innerKey] instanceof EventEmitter &&
                ((/** @type {?} */ (this.innerComponent[innerKey]))).observers.length > 0)
                .forEach(innerKey => (getBindIOMetadata(this.innerComponent).asInner.manualOutputs[innerKey] = ((/** @type {?} */ (this.innerComponent[innerKey]))).observers.length));
        }
    }
    /**
     * @return {?}
     */
    debugIsActive() {
        return (this._ngxBindIOConfig.debug ||
            (this.bindIO && this.bindIO.debug) ||
            (localStorage && localStorage.getItem('debug_ngx-bind-io') === 'true')); // todo: remove on stable release
    }
}
BindIODirective.decorators = [
    { type: Directive, args: [{
                selector: '[bindIO]'
            },] }
];
/** @nocollapse */
BindIODirective.ctorParameters = () => [
    { type: ViewContainerRef },
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: NgxBindInputsService },
    { type: NgxBindOutputsService },
    { type: NgxBindIODebugService },
    { type: ChangeDetectorRef }
];
BindIODirective.propDecorators = {
    bindIO: [{ type: Input }],
    excludeIO: [{ type: Input }],
    includeIO: [{ type: Input }],
    excludeInputs: [{ type: Input }],
    includeInputs: [{ type: Input }],
    excludeOutputs: [{ type: Input }],
    includeOutputs: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindOutputsDirective {
    /**
     * @param {?} viewContainerRef
     * @param {?} _ngxBindIOConfig
     * @param {?} _ngxBindOutputsService
     * @param {?} _ngxBindIODebugService
     * @param {?} _ref
     */
    constructor(viewContainerRef, _ngxBindIOConfig, _ngxBindOutputsService, _ngxBindIODebugService, _ref) {
        this.viewContainerRef = viewContainerRef;
        this._ngxBindIOConfig = _ngxBindIOConfig;
        this._ngxBindOutputsService = _ngxBindOutputsService;
        this._ngxBindIODebugService = _ngxBindIODebugService;
        this._ref = _ref;
        this.excludeOutputs = [];
        this.includeOutputs = [];
        this.usedOutputs = {};
        this.destroyed$ = new Subject();
    }
    /**
     * @param {?} simpleChanges
     * @return {?}
     */
    ngOnChanges(simpleChanges) {
        this.detectComponents();
    }
    /**
     * @return {?}
     */
    ngOnInit() {
        this.detectComponents();
        this.bindAll();
    }
    /**
     * @return {?}
     */
    bindAll() {
        if (this.outputs === undefined) {
            this.outputs = this._ngxBindOutputsService.getOutputs(this);
            this._ngxBindOutputsService.bindOutputs(this);
            this._ngxBindIODebugService.showDebugInfo(this, this.debugIsActive());
        }
    }
    /**
     * @return {?}
     */
    ngOnDestroy() {
        this.destroyed$.next(true);
        this.destroyed$.complete();
    }
    /**
     * @return {?}
     */
    detectComponents() {
        if (!this.innerComponent && !this.hostComponent) {
            this.innerComponent = this.viewContainerRef['_data'].componentView.component;
            this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.context;
            if (this.hostComponent.$implicit !== undefined) {
                this.hostComponent = ((/** @type {?} */ (this.viewContainerRef)))._view.component;
            }
            getBindIOMetadata(this.innerComponent).asInner.manualOutputs = {};
            Object.keys(this.innerComponent)
                .filter(innerKey => this.innerComponent[innerKey] instanceof EventEmitter &&
                ((/** @type {?} */ (this.innerComponent[innerKey]))).observers.length > 0)
                .forEach(innerKey => (getBindIOMetadata(this.innerComponent).asInner.manualOutputs[innerKey] = ((/** @type {?} */ (this.innerComponent[innerKey]))).observers.length));
        }
    }
    /**
     * @return {?}
     */
    debugIsActive() {
        return (this._ngxBindIOConfig.debug ||
            (this.bindOutputs && this.bindOutputs.debug) ||
            (localStorage && localStorage.getItem('debug_ngx-bind-io') === 'true')); // todo: remove on stable release
    }
}
BindOutputsDirective.decorators = [
    { type: Directive, args: [{
                selector: '[bindOutputs]'
            },] }
];
/** @nocollapse */
BindOutputsDirective.ctorParameters = () => [
    { type: ViewContainerRef },
    { type: undefined, decorators: [{ type: Inject, args: [NGX_BIND_IO_CONFIG,] }] },
    { type: NgxBindOutputsService },
    { type: NgxBindIODebugService },
    { type: ChangeDetectorRef }
];
BindOutputsDirective.propDecorators = {
    bindOutputs: [{ type: Input }],
    excludeOutputs: [{ type: Input }],
    includeOutputs: [{ type: Input }]
};

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class NgxBindIOModule {
    /**
     * @param {?=} options
     * @return {?}
     */
    static forRoot(options) {
        return {
            ngModule: NgxBindIOModule,
            providers: [
                {
                    provide: NGX_BIND_IO_CONFIG,
                    useValue: {
                        debug: options ? options.debug : defaultNgxBindIOConfig.debug
                    }
                },
                NgxBindIODebugService,
                NgxBindInputsService,
                NgxBindOutputsService
            ]
        };
    }
}
NgxBindIOModule.decorators = [
    { type: NgModule, args: [{
                imports: [CommonModule],
                declarations: [BindInputsDirective, BindOutputsDirective, BindIODirective],
                exports: [BindInputsDirective, BindOutputsDirective, BindIODirective]
            },] }
];

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */
class BindIoInnerLifecycle {
    /**
     * @param {?} simpleChanges
     * @return {?}
     */
    ngOnChanges(simpleChanges) {
        if (!getBindIOMetadata(this).asInner.manualInputs) {
            getBindIOMetadata(this).asInner.manualInputs = {};
            if (simpleChanges) {
                Object.keys(simpleChanges).forEach(innerKey => (getBindIOMetadata(this).asInner.manualInputs[innerKey] = 1));
            }
        }
        if (typeof ((/** @type {?} */ (this))).__originalNgOnChanges__ === 'function') {
            ((/** @type {?} */ (this))).__originalNgOnChanges__(simpleChanges);
        }
    }
}
/**
 * @return {?}
 */
function BindIoInner() {
    return function (target) {
        if (!target.prototype.__originalNgOnChanges__) {
            /** @type {?} */
            const bindIoInnerLifecycle = new BindIoInnerLifecycle();
            target.prototype.__originalNgOnChanges__ = target.prototype.ngOnChanges;
            target.prototype.ngOnChanges = bindIoInnerLifecycle.ngOnChanges;
        }
    };
}

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

/**
 * @fileoverview added by tsickle
 * @suppress {checkTypes,extraRequire,missingReturn,unusedPrivateMembers,uselessCode} checked by tsc
 */

export { defaultNgxBindIOConfig, NGX_BIND_IO_CONFIG, NgxBindIOModule, BindIoInner, BindIoInnerLifecycle, BindInputsDirective, BindIODirective, BindOutputsDirective, NgxBindInputsService, NgxBindIODebugService, NgxBindOutputsService, getBindIOMetadata, collectKeys, removeKeysUsedInAttributes, removeKeysManualBindedOutputs, removeKeysManualBindedInputs, redefineSimpleProperty, redefineAccessorProperty, getPropDescriptor, isFunction, applyMixins, Mixin };

//# sourceMappingURL=ngx-bind-io.js.map